from django.db import connection
from django.db.models import Field


def get_field_name(field: Field) -> str:
    if hasattr(field, "name"):
        field_name = field.name
        if hasattr(field, "primary_key") and field.primary_key:
            field_name = f"{field_name} (pk)"
        return field_name

    return ""


def get_field_column(field: Field) -> str:
    if hasattr(field, "column"):
        return str(field.column)

    return ""


def get_field_verbose_name(field: Field) -> str:
    if hasattr(field, "verbose_name"):
        return str(field.verbose_name)

    return ""


def get_field_db_type(field: Field) -> str:
    db_type = "None"
    if hasattr(field, "db_type"):
        try:
            db_type = str(field.db_type(connection=connection))
        except TypeError:
            db_type = " TypeError (db_type)"
        except AttributeError:
            db_type = " AttributeError (db_type)"

        if db_type == "None" and hasattr(field, "through"):
            db_type = f"through {field.through._meta.label}"

        if db_type == "None" and (
            hasattr(field, "remote_field") and field.remote_field is not None and hasattr(field.remote_field, "through")
        ):
            db_type = f"through {field.remote_field.through._meta.label}"
        return db_type

    return db_type


def get_related_name(field: Field) -> str:
    if hasattr(field, "related_name") and field.related_name is not None:
        field_related_name = str(field.related_name)
        if "+" in field_related_name:
            return f"{field_related_name} (no reverse relation allowed)"
        return field_related_name

    elif (
        hasattr(field, "remote_field")
        and field.remote_field is not None
        and hasattr(field.remote_field, "related_name")
    ):
        return str(field.remote_field.related_name)

    return "(None)"


def get_field_name_on_reverse_model(field: Field) -> str:
    if hasattr(field, "field") and field.field is not None:
        return str(field.field.name)

    return ""


def get_field_type_on_reverse_model(field: Field) -> str:
    if hasattr(field, "field"):
        return str(field.field.get_internal_type())

    return ""


def get_field_type(field: Field) -> str:
    if hasattr(field, "__class__"):
        return str(field.__class__.__name__)
    elif hasattr(field, "get_internal_type"):
        return str(field.get_internal_type())

    return ""


def get_related_model(field: Field) -> str:
    if hasattr(field, "related_model") and field.related_model is not None:
        return str(field.related_model._meta.label)

    return ""



from dataclasses import astuple, dataclass


@dataclass
class ModelInfo:
    """Class for keeping track of fields that are relations."""

    @dataclass
    class Annotated:
        title: str = ""
        value: str = ""

    model_name: Annotated = Annotated(title="Model Name", value="")
    verbose_name: str = Annotated(title="Verbose Name", value="")
    docstring: str = Annotated(title="Docstring", value="")
    is_abstract: str = Annotated(title="Is Abstract", value="")
    is_proxy: str = Annotated(title="Is Proxy", value="")
    is_managed: str = Annotated(title="Is Managed", value="")
    database_table: str = Annotated(title="Database Tables", value="")
    base_manager: str = Annotated(title="Base Manager", value="")
    default_manager: str = Annotated(title="Default Manager", value="")
    file: str = Annotated(title="File", value="")
    line_number: str = Annotated(title="Starting Line Number", value="")

    def render_rows(self, row_count: int) -> list:
        rows = []
        for row_num, entry in enumerate(self.__dataclass_fields__):
            if row_num <= row_count:
                field = getattr(self, entry)
                title = field.title
                value = field.value
                rows.append((title, value))
        return rows

    def render_simple_rows(self) -> list:
        """Renders a single row"""
        return [self.model_name.title, self.model_name.value]

    def __str__(self):
        return self.model_name.value


@dataclass
class BaseInfo:
    """Base class for keeping track of model fields and methods."""

    name: str = ""

    def render_row(self, column_count: int) -> list:
        return list(astuple(self))[:column_count]

    def render_simple_row(self) -> list:
        return [
            self.name,
        ]

    def __str__(self):
        return self.name


@dataclass
class FieldRelation(BaseInfo):
    """Class for keeping track of fields that are relations."""

    field_type: str = ""
    field_column: str = ""
    field_db_type: str = ""
    related_model: str = ""
    related_name: str = ""


@dataclass
class FieldReverseRelation(BaseInfo):
    """Class for keeping track of fields that are relations."""

    field_type: str = ""
    field_db_type: str = ""
    related_model: str = ""
    field_name_on_related_model: str = ""
    field_type_on_related_model: str = ""


@dataclass
class FieldOther(BaseInfo):
    """Class for keeping track of fields that are relations."""

    field_type: str = ""
    field_column: str = ""
    field_db_type: str = ""
    field_verbose_name: str = ""


@dataclass
class Method(BaseInfo):
    """Class for keeping track of dunder methods."""

    method_signature: str = ""
    method_docstring: str = ""
    method_file: str = ""
    method_line_number: str = ""



import inspect


def get_method_signature(method: str, model: type, verbosity: int) -> str:
    if verbosity > 1:

        try:
            return str(inspect.signature(getattr(model, method)))
        except ValueError:
            return "ValueError no signature found"
        except TypeError as e:
            return f"TypeError {e}"
        except Exception as e:
            return f"{e}"

    return ""


def get_method_docstring(method: str, model: type) -> str:
    try:
        return inspect.getdoc(getattr(model, method))
    except:
        pass

    return ""


def get_method_file(method: str, model: type) -> str:
    try:
        return inspect.getfile(getattr(model, method))
    except:
        pass

    return ""


def get_method_line_number(method: str, model: type) -> str:
    try:
        file_info = inspect.getsourcelines(getattr(model, method))
        if len(file_info) > 0:
            return str(file_info[-1])
    except:
        pass

    return ""



import inspect

from django.db.models import Model


def get_model_name(model: Model) -> str:
    if hasattr(model, "__name__"):
        return str(model.__name__)

    return ""


def get_model_verbose_name(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "verbose_name"):
        return str(model._meta.verbose_name)

    return ""


def get_model_database_table(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "db_table"):
        return str(model._meta.db_table)

    return ""


def get_model_is_abstract(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "abstract"):
        return str(model._meta.abstract)

    return ""


def get_model_is_managed(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "managed"):
        return str(model._meta.managed)

    return ""


def get_model_is_proxy(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "proxy"):
        return str(model._meta.proxy)

    return ""


def get_model_base_manager(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "base_manager_name"):
        return str(model._meta.base_manager_name)

    return ""


def get_model_default_manager(model: Model) -> str:
    if hasattr(model, "_meta") and hasattr(model._meta, "default_manager_name"):
        return str(model._meta.default_manager_name)

    return ""


def get_model_docstring(model: Model) -> str:
    try:
        return str(inspect.getdoc(model))
    except:
        pass

    return ""


def get_model_file(model: Model) -> str:
    try:
        return str(inspect.getfile(model))
    except:
        pass

    return ""


def get_model_line_number(model: Model) -> str:
    try:
        file_info = inspect.getsourcelines(model)
        if len(file_info) > 0:
            return str(file_info[-1])
    except:
        pass

    return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from django_model_info/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("django_model_info", "__init__.py")


if sys.argv[-1] == 'publish':
    try:
        import wheel
        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

with open("HISTORY.md", "r", encoding="utf-8") as fh:
    history = fh.read()

requirements = open('requirements.txt').readlines()

setup(
    name='django-model-info',
    version=version,
    description="""A Django Management Command for displaying details about your project's models""",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    author='Jack Linke',
    author_email='jack@watervize.com',
    url='https://github.com/jacklinke/django-model-info',
    packages=[
        'django_model_info',
    ],
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='django-model-info',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Plugins',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License  ',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Documentation',
        'Topic :: Utilities',
    ],
)



import os
from pathlib import Path

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.base import BaseCommand, CommandParser, DjangoHelpFormatter
from django.db.models import Model
from rich.align import Align
from rich.bar import Bar
from rich.console import Console
from rich.padding import Padding
from rich.style import Style
from rich.table import Table

from ._field_attr_utils import (
    get_field_column,
    get_field_db_type,
    get_field_name,
    get_field_name_on_reverse_model,
    get_field_type,
    get_field_type_on_reverse_model,
    get_field_verbose_name,
    get_related_model,
    get_related_name,
)
from ._info_classes import FieldOther, FieldRelation, FieldReverseRelation, Method, ModelInfo
from ._method_attr_utils import get_method_docstring, get_method_file, get_method_line_number, get_method_signature
from ._model_attr_utils import (
    get_model_base_manager,
    get_model_database_table,
    get_model_default_manager,
    get_model_docstring,
    get_model_file,
    get_model_is_abstract,
    get_model_is_managed,
    get_model_is_proxy,
    get_model_line_number,
    get_model_name,
    get_model_verbose_name,
)

console = Console(record=True)


DEFAULT_DJANGO_METHODS = (
    "_check_column_name_clashes",
    "_check_constraints",
    "_check_default_pk",
    "_check_field_name_clashes",
    "_check_fields",
    "_check_id_field",
    "_check_index_together",
    "_check_indexes",
    "_check_local_fields",
    "_check_long_column_names",
    "_check_m2m_through_same_relationship",
    "_check_managers",
    "_check_model",
    "_check_model_name_db_lookup_clashes",
    "_check_ordering",
    "_check_property_name_related_field_accessor_clashes",
    "_check_single_primary_key",
    "_check_swappable",
    "_check_unique_together",
    "_do_insert",
    "_do_update",
    "_get_expr_references",
    "_get_FIELD_display",
    "_get_next_or_previous_by_FIELD",
    "_get_next_or_previous_in_order",
    "_get_pk_val",
    "_get_unique_checks",
    "_meta",
    "_perform_date_checks",
    "_perform_unique_checks",
    "_prepare_related_fields_for_save",
    "_save_parents",
    "_save_table",
    "_set_pk_val",
    "check",
    "clean",
    "clean_fields",
    "date_error_message",
    "delete",
    "from_db",
    "full_clean",
    "get_absolute_url",
    "get_deferred_fields",
    "prepare_database_save",
    "refresh_from_db",
    "save",
    "save_base",
    "serializable_value",
    "unique_error_message",
    "validate_unique",
)


class Command(BaseCommand):
    """
    A management command which lists models within your project, and optionally, details about model fields and methods

    Verbosity outputs:
        0   Model names only - Convenient when you just need a list of all your project's models in one place
        1   Model names, field names, and non-dunder/common method names
        2 * Model names, field names & details, and non-dunder/common method names & details
        3   Model names, field names & details, and all method names & full details

        * Verbosity of 2 is default
    """

    help = "List out the fields and methods for each model"

    def create_parser(self, prog_name, subcommand, **kwargs):
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this command.

        Reimplemented to allow new default verbosity of 2
        """
        parser = CommandParser(
            prog="%s %s" % (os.path.basename(prog_name), subcommand),
            description=self.help or None,
            formatter_class=DjangoHelpFormatter,
            missing_args_message=getattr(self, "missing_args_message", None),
            called_from_command_line=getattr(self, "_called_from_command_line", None),
            **kwargs,
        )
        parser.add_argument("--version", action="version", version=self.get_version())
        parser.add_argument(
            "--settings",
            help=(
                "The Python path to a settings module, e.g. "
                '"myproject.settings.main". If this isn\'t provided, the '
                "DJANGO_SETTINGS_MODULE environment variable will be used."
            ),
        )
        parser.add_argument(
            "--pythonpath",
            help='A directory to add to the Python path, e.g. "/home/djangoprojects/myproject".',
        )
        parser.add_argument("--traceback", action="store_true", help="Raise on CommandError exceptions")
        parser.add_argument(
            "--no-color",
            action="store_true",
            help="Don't colorize the command output.",
        )
        parser.add_argument(
            "--force-color",
            action="store_true",
            help="Force colorization of the command output.",
        )
        if self.requires_system_checks:
            parser.add_argument(
                "--skip-checks",
                action="store_true",
                help="Skip system checks.",
            )
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "-v",
            "--verbosity",
            default=2,
            type=int,
            choices=[0, 1, 2, 3],
            help="Verbosity level: "
            "0 Model names only - Convenient when you just need a list of all your project's models in one place, "
            "1 Model names + field names +non-dunder/common method names, "
            "2 (default) Model names + field names & details + non-dunder/common method names & details, "
            "3 Model names + field names & details + all method names & details.",
        )
        parser.add_argument(
            "-e",
            "--export",
            nargs="?",
            type=str,
            default=None,
            help="Filename to export. The filename must have a file extension of `.txt`, `.html`, or `htm`",
        )
        parser.add_argument(
            "-f",
            "--filter",
            nargs="+",
            type=str,
            default=None,
            help="Provide one or more apps or models, to which the results will be limited. "
            "Input should be in the form `appname` or `appname.Modelname`.",
        )

    def model_info(self, options) -> tuple:
        section_style = Style(color="green", bold=True, underline=True)
        subsection_style = Style(color="green", bold=True)

        def get_options():
            VERBOSITY = options.get("verbosity", None)
            if VERBOSITY is None:
                VERBOSITY = (
                    getattr(settings, "MODEL_INFO_VERBOSITY", 2)
                    if type(getattr(settings, "MODEL_INFO_VERBOSITY", 2)) is int
                    else 2
                )

            FILTER = options.get("filter", None)
            if FILTER is None:
                FILTER = (
                    getattr(settings, "MODEL_INFO_FILTER", None)
                    if type(getattr(settings, "MODEL_INFO_FILTER", None)) is list
                    else None
                )
            FILENAME = (
                options.get("export")
                if options.get("export", None) is not None and type(options.get("export", None)) is str
                else None
            )

            return VERBOSITY, FILTER, FILENAME

        VERBOSITY, FILTER, FILENAME = get_options()

        def build_model_objects(model) -> ModelInfo:
            """
            Given a model, returns a ModelInfo object
            """

            new_model = ModelInfo()
            new_model.model_name.value = get_model_name(model)
            new_model.verbose_name.value = get_model_verbose_name(model)
            new_model.docstring.value = get_model_docstring(model)
            new_model.is_abstract.value = get_model_is_abstract(model)
            new_model.is_proxy.value = get_model_is_proxy(model)
            new_model.is_managed.value = get_model_is_managed(model)
            new_model.database_table.value = get_model_database_table(model)
            new_model.base_manager.value = get_model_base_manager(model)
            new_model.default_manager.value = get_model_default_manager(model)
            new_model.file.value = get_model_file(model)
            new_model.line_number.value = get_model_line_number(model)

            return new_model

        def build_field_objects(field_list: list) -> tuple:
            """
            Given a list of model fields, returns a tuple of FieldRelation,
            FieldReverseRelation, and FieldOther object lists
            """
            fields_relation = []
            fields_reverse_relation = []
            fields_other = []

            for field in field_list:
                # Identify the kind of field this is, and build associated object

                if hasattr(field, "related_model") and field.related_model is not None:
                    if "reverse_related" in field.__class__.__module__.__str__():
                        # Build a FieldReverseRelation object
                        new_field = FieldReverseRelation()
                        new_field.name = get_related_name(field)
                        new_field.field_type = get_field_type(field)
                        new_field.field_db_type = get_field_db_type(field)
                        new_field.related_model = get_related_model(field)
                        new_field.field_name_on_related_model = get_field_name_on_reverse_model(field)
                        new_field.field_type_on_related_model = get_field_type_on_reverse_model(field)

                        fields_reverse_relation.append(new_field)
                    else:
                        # Build a FieldRelation object
                        new_field = FieldRelation()
                        new_field.name = get_field_name(field)
                        new_field.field_type = get_field_type(field)
                        new_field.field_column = get_field_column(field)
                        new_field.field_db_type = get_field_db_type(field)
                        new_field.related_model = get_related_model(field)
                        new_field.related_name = get_related_name(field)

                        fields_relation.append(new_field)
                else:
                    # Build a FieldOther object
                    new_field = FieldOther()
                    new_field.name = get_field_name(field)
                    new_field.field_type = get_field_type(field)
                    new_field.field_column = get_field_column(field)
                    new_field.field_db_type = get_field_db_type(field)
                    new_field.field_verbose_name = get_field_verbose_name(field)

                    fields_other.append(new_field)

            return (
                fields_relation,
                fields_reverse_relation,
                fields_other,
            )

        def build_method_objects(method_list: list, model: Model) -> tuple:
            """
            Given a list of model methods, returns a tuple of MethodCommonDjango,
            MethodDunder, MethodOther, MethodOtherPrivate object lists
            """

            method_dunder = []
            method_common_django = []
            method_other_private = []
            method_other = []

            for method in method_list:
                # Build the object, and assign to the correct list
                new_method = Method()

                new_method.name = method
                if VERBOSITY > 1:
                    new_method.method_signature = get_method_signature(method, model, VERBOSITY)
                if VERBOSITY > 2:
                    new_method.method_docstring = get_method_docstring(method, model)
                    new_method.method_file = get_method_file(method, model)
                    new_method.method_line_number = get_method_line_number(method, model)

                if method.startswith("__") and method.endswith("__"):
                    # Dunder methods
                    method_dunder.append(new_method)

                elif method in DEFAULT_DJANGO_METHODS:
                    # Common Django methods
                    method_common_django.append(new_method)

                elif method.startswith("_"):
                    # Other Private methods
                    method_other_private.append(new_method)

                else:
                    # Other methods
                    method_other.append(new_method)

            return (
                method_dunder,
                method_common_django,
                method_other_private,
                method_other,
            )

        def _fill_table(info_table: Table, info_object_list: list or None, info_type: type, column_count: int):
            """
            Given a rich table, a list of
            """
            if isinstance(info_object_list, list) and all(isinstance(row, info_type) for row in info_object_list):
                sorted_field_object_list = sorted(info_object_list, key=lambda x: x.name)
                for row in sorted_field_object_list:

                    if VERBOSITY >= 2:
                        info_table.add_row(*row.render_row(column_count=column_count))
                    else:
                        info_table.add_row(*row.render_simple_row())

            else:
                info_table.add_row("none")

            return info_table

        def _print_table(table):
            console.print(Padding(table, (1, 0, 0, 8)))

        def render_model_table(info_object_list: list or None, info_type: type):
            """Provided a list of FieldRelation objects, prints the resulting sorted table to console"""
            model_table = Table(title="Model Details")
            row_count = 2
            if VERBOSITY > 1:
                row_count = 5
            if VERBOSITY > 2:
                row_count = 11

            model_table.add_column("Key", justify="left", style="blue", no_wrap=True)
            model_table.add_column("Value", justify="left", style="magenta")

            if isinstance(info_object_list, ModelInfo):
                for row in info_object_list.render_rows(row_count):
                    new_row = tuple(row)
                    model_table.add_row(new_row[0], new_row[1])

            else:
                model_table.add_row("none")

            _print_table(model_table)

        def render_field_relations_table(info_object_list: list or None, info_type: type):
            """Provided a list of FieldRelation objects, prints the resulting sorted table to console"""
            field_table = Table(title="Relations")
            column_count = 1

            field_table.add_column("Field Name", justify="left", style="yellow", no_wrap=True)
            if VERBOSITY >= 2:
                column_count = 6
                field_table.add_column("Field Type", justify="left", style="magenta")
                field_table.add_column("Database Column", justify="left", style="magenta")
                field_table.add_column("Database Type", justify="left", style="magenta")
                field_table.add_column("Related Model", justify="right", style="dark_red")
                field_table.add_column("Related Name", justify="right", style="dark_red")

            field_table = _fill_table(field_table, info_object_list, info_type, column_count)
            _print_table(field_table)

        def render_field_reverse_relations_table(info_object_list: list or None, info_type: type):
            """Provided a list of FieldReverseRelation objects, prints the resulting sorted table to console"""
            field_table = Table(title="Reverse Relations")
            column_count = 1

            field_table.add_column("Related Name", justify="left", style="yellow", no_wrap=True)
            if VERBOSITY >= 2:
                column_count = 7
                field_table.add_column("Field Type", justify="left", style="magenta")
                field_table.add_column("Database Type", justify="left", style="magenta")
                field_table.add_column("Related Model", justify="right", style="dark_red")
                field_table.add_column("Field Name on Related Model", justify="left", style="dark_red")
                field_table.add_column("Field Type on Related Model", justify="left", style="dark_red")

            field_table = _fill_table(field_table, info_object_list, info_type, column_count)
            _print_table(field_table)

        def render_field_others_table(info_object_list: list or None, info_type: type):
            """Provided a list of FieldOther objects, prints the resulting sorted table to console"""
            field_table = Table(title="Other Fields")
            column_count = 1

            field_table.add_column("Field Name", justify="left", style="yellow", no_wrap=True)
            if VERBOSITY >= 2:
                column_count = 6
                field_table.add_column("Field Type", justify="left", style="magenta")
                field_table.add_column("Database Column", justify="left", style="magenta")
                field_table.add_column("Database Type", justify="left", style="magenta")
                field_table.add_column("Verbose Name", justify="left", style="white")

            field_table = _fill_table(field_table, info_object_list, info_type, column_count)
            _print_table(field_table)

        def render_method_table(info_object_list: list or None, info_type: str):
            """Provided a list of Method objects, prints the resulting sorted table to console"""
            method_table = Table(title=info_type)
            column_count = 1

            method_table.add_column("Method Name", justify="left", style="cyan", no_wrap=True)
            if VERBOSITY > 1:
                column_count = 2
                method_table.add_column("Signature", justify="left", style="magenta")
            if VERBOSITY > 2:
                column_count = 5
                method_table.add_column("Docstring", justify="left", style="magenta")
                method_table.add_column("File", justify="left", style="magenta")
                method_table.add_column("Line Number", justify="left", style="magenta")

            method_table = _fill_table(method_table, info_object_list, Method, column_count)
            _print_table(method_table)

        def get_model_list():
            if FILTER is not None:
                model_list = []
                for filter_item in FILTER:
                    if filter_item.count(".") == 0:
                        # Get the models and add to the list
                        # model_list.append(django_apps.get_app_config(filter_item).get_models())
                        try:
                            app_models = [x for x in django_apps.get_app_config(filter_item).get_models()]
                        except LookupError as e:
                            print(f"Error while looking up `{filter_item}`: {e}")
                        else:
                            model_list.extend(app_models)
                    elif filter_item.count(".") == 1:
                        # Add to the model list
                        try:
                            filter_model = django_apps.get_model(filter_item)
                        except LookupError as e:
                            print(f"Error while looking up `{filter_item}`: {e}")
                        else:
                            model_list.append(filter_model)
            else:
                model_list = sorted(
                    django_apps.get_models(), key=lambda x: (x._meta.app_label, x._meta.object_name), reverse=False
                )
            return model_list

        model_list = get_model_list()

        for model in model_list:
            if VERBOSITY > 0:
                console.print(Padding("", (1, 0, 0, 0)))
                console.print(Padding("", (0, 0, 0, 0), style=section_style))
                console.print(Padding("", (0, 0, 0, 0)))
            console.print(f"{model._meta.label}", style=section_style)

            if VERBOSITY > 0:

                def process_model():
                    build_model_objects(model)
                    model_info = build_model_objects(model)
                    render_model_table(model_info, list)

                process_model()

                def process_fields():
                    console.print(Padding("Fields:", (1, 0, 0, 4), style=subsection_style))

                    field_list = model._meta.get_fields(include_hidden=True)

                    fields_relation, fields_reverse_relation, fields_other = build_field_objects(field_list)

                    render_field_relations_table(fields_relation, FieldRelation)
                    render_field_reverse_relations_table(fields_reverse_relation, FieldReverseRelation)
                    render_field_others_table(fields_other, FieldOther)

                process_fields()

                def get_clean_method_list():
                    """
                    Remove any potential method names that start with an uppercase character, are blank, or not callable
                    """
                    return [
                        method_name
                        for method_name in dir(model)
                        if method_name is not None
                        and not method_name == ""
                        and not method_name[0].isupper()
                        and hasattr(model, method_name)
                        and callable(getattr(model, method_name))
                    ]

                method_list = get_clean_method_list()

                def process_methods():
                    if VERBOSITY == 3:
                        console.print(Padding("Methods (all):", (1, 0, 0, 4), style=subsection_style))
                    else:
                        console.print(Padding("Methods (non-private/internal):", (1, 0, 0, 4), style=subsection_style))

                    method_dunder, method_common_django, method_other_private, method_other = build_method_objects(
                        method_list, model
                    )

                    if VERBOSITY > 1:
                        render_method_table(method_dunder, "Dunder Methods")
                        render_method_table(method_common_django, "Common Django Methods")
                    render_method_table(method_other_private, "Other Private methods")
                    render_method_table(method_other, "Other Methods")

                process_methods()

                self.stdout.write("\n")

        console.print(f"\nTotal Models Listed: {len(model_list)}\n", style=section_style)
        console.print(Align(Bar(size=0.1, begin=0.0, end=0.0, width=100), align="center"), style="red")

        def process_export():
            """If a FILENAME was provided in options, try to save the appropriate type of file"""
            if FILENAME is not None:
                extension = Path(FILENAME).suffixes

                if len(extension) > 0:
                    if any(x in extension[-1] for x in ["htm", "html"]):
                        console.save_html(path=FILENAME)
                        # Using print() to avoid exporting following line
                        print(f"Saved as {FILENAME}")
                    elif "txt" in extension[-1]:
                        console.save_text(path=FILENAME)
                        # Using print() to avoid exporting following line
                        print(f"Saved as {FILENAME}")

        process_export()

    def handle(self, *args, **options):
        self.model_info(options)
